#include <iostream>
#include "superresolution.h"

using namespace std;

//=============================================================================
// superresolution
//=============================================================================
// Deze functie implementeert het superresolutiealgoritme
//=============================================================================
// Argumenten:
// - input: invoer PPM sequentie
// - refnr: nummer van het beeld dat als referentie gebruikt wordt
// - factor: het aantal keren dat het beeld vergroot wordt (breedte en hoogte)
// - type: 
//=============================================================================
frame superresolution(ppmsequence& input,int refnr,int factor,byte type,double sigma,double lambda,double step,int iterations) {

	// Enkele variabelen die verder gebruikt worden
	frame timage;
	int j2,i2;
	double halfwidth = (double)(input.width - 1)/2.0;
	double halfheight = (double)(input.height - 1)/2.0;
	int maxiter = 100;
	int bufferlength = MIN((int)(input.frames),15);

	//========================================================================
	// STAP 1: initializatie
	//========================================================================
	// Maak een frame aan voor het eindresultaat, dat (factor) keer groter is
	frame result((int)(input.height)*factor,(int)(input.width)*factor);
	// Selecteer een referentiebeeld
	frame image = input.get_frame(refnr);
	// Voor de fijne registratie, wordt ook een B-spline kubische interpolatie gedaan 
	// op het referentiebeeld
	frame highimage = interpolation_bspline_cubic(image,factor*image.height,factor*image.width);

	// Andere hulpvariabelen
	cube<byte_rgb> stock(result.height,result.width,bufferlength);
	bytematrix mask(result.height,result.width);
	intmatrix countmask(result.height,result.width);

	// Voor het eindresultaat wordt vertrokken van een leeg beeld
	for (int j = 0; j < result.height; j++) {
		for (int i = 0; i < result.width; i++) {
			result(j,i) = black;
			mask(j,i) = 0;
			countmask(j,i) = 0;
			for (int k = 0; k < stock.length; k++) {
				stock(j,i,k) = black;
			}
		}
	}

	//========================================================================
	// STAP 2: registratie, via translationeel model
	// De beelden worden telkens geregistreerd ten opzichte van het 
	// referentiebeeld
	//========================================================================
	cout << "Registratie..." << endl;
	rgb pixel;
	int searchwindow = 10;	// Grootte van het zoekvenster

	/* tip: gebruik:
	  - input.frames: om het aantal invoerframes te kennen
	  - searchwindow: voor de grootte van het zoekvenster 
	  - timage = input.get_frame(nr) om een bepaald invoerframe naar timage te kopiren
	  - image: is het referentiebeeld
	  - factor: de uitvergrotingsfactor
	  - highimage: een referentie hogeresolutiebeeld
	  - stock(j,i,k): om de geregistreerde beelden in op te slaan
			j: rij-index in het beeld
			i: kolom-index in het beeld
			k: de derde dimensie bevat een rij met intensiteiten van geregistreerde 
			   beelden. k is niet noodzakelijk gelijk aan nr (frame nummer), aangezien
			   sommige beelden geen data bevatten voor sommige posities in het 
			   hogerresolutiebeeld. countmask(j,i) bevat het aantal intensiteiten 
			   dat is opgemeten voor positive (j,i). In de fusiestap (zie lager)
			   worden vervolgens "stock" en "countmask" gebruikt om tot n enkele
			   schatting te komen van het hogerresolutiebeeld.
	*/

	/* Implementeer de beeldregistratie hier */

	//========================================================================
	// STAP 3: fusie van de beelden
	//========================================================================
	cout << "Robuuste fusie van de beelden..." << endl;
	doublevector buffer_luminance;

	double lum;
	for (int j = 0; j < result.height; j++) {
		for (int i = 0; i < result.width; i++) {
			// median scheme
			if (countmask(j,i) > 0) {
				i2 = countmask(j,i) >> 1;
				buffer_luminance = doublevector(countmask(j,i));
				for (int k = 0; k < countmask(j,i); k++) {
					pixel = rgb((double)(stock(j,i,k).red),(double)(stock(j,i,k).green),(double)(stock(j,i,k).blue));
					buffer_luminance(k) = rgb2grey(pixel);
				}
				for (int k = 0; k <= i2; k++) {
					j2 = find_vector_min(buffer_luminance,lum);
					buffer_luminance(j2) = 256.0;
				}
				result(j,i) = rgb((double)(stock(j,i,j2).red),(double)(stock(j,i,j2).green),(double)(stock(j,i,j2).blue));
			} else {
				result(j,i) = highimage(j,i);
			}
		}
	}

	//========================================================================
	// STAP 4: gezamelijke ruisonderdrukking+verscherping
	//========================================================================
	switch(type) {
		// Tikhonov
		case 0: cout << "Iteratieve L2 + Tikhonov deblurring/denoising..." << endl;
			result = filter_robust_regularization(result,0,0,sigma,lambda,iterations,step);
			break;
			// Total Variation
		case 1: cout << "Iteratieve L1 + Total Variation deblurring/denoising..." << endl;
			result = filter_robust_regularization(result,1,1,sigma,lambda,iterations,step);
			break;
			// Bilateral Total Variation
		case 2: cout << "Iteratieve L1 + Bilateral Total Variation deblurring/denoising..." << endl;
			result = filter_robust_regularization(result,1,2,sigma,lambda,iterations,step);
			break;
			// Lorentzian
		case 3: cout << "Iteratieve L1 + Lorentzian deblurring/denoising..." << endl;
			result = filter_robust_regularization(result,1,3,sigma,lambda,iterations,step);
			break;
			// Tikhonov
		default: cout << "Iteratieve L2 + Tikhonov deblurring/denoising..." << endl;
			result = filter_robust_regularization(result,0,0,sigma,lambda,iterations,step);
			break;
	}
	return result;
}
